﻿using WpfOsi.BaseClasses;
using WpfOsi.Helper.IO;
using WpfOsi.Model;
using System;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;

namespace WpfOsi.Context
{
    [Serializable]
    public class OsiContext:NotifyBase, IDisposable
    {
        private readonly string _xmlpath;

        private const string LockFileName = "pessimisticlock.loc";
        private DateTime _lastSyncTime;

        private FileSystemWatcher _xmlFileWatcher;
        private XmlContext CurrentContext { get; set; }

        internal event DataFileWasChangedEventHandler DataFileWasChanged;
        internal delegate void DataFileWasChangedEventHandler();
        internal event ErrorMessageSentEventHandler ErrorMessageSent;
        internal delegate void ErrorMessageSentEventHandler(string message);

        public OsiContext()
        {
            if (String.IsNullOrEmpty(Properties.Settings.Default.dbFileXmlPath)) throw new Exception("Der Pfad zum XML-Datenfile muss in den Settings hinterlegt sein.");
            if (!File.Exists(Properties.Settings.Default.dbFileXmlPath)) throw new Exception("Der Pfad zur XML-Datendatei aus den Settings existiert nicht.");
            _xmlpath = Properties.Settings.Default.dbFileXmlPath;
            InitOsiContext();
        }
        public OsiContext(string xmlpath, bool savePathToSettings = false)
        {
            if (!Directory.Exists(Path.GetDirectoryName(xmlpath))) throw new Exception("Der Pfad zum XML-Datendateiordner existiert nicht.");
            _xmlpath = xmlpath;
            if (savePathToSettings) Properties.Settings.Default.dbFileXmlPath = _xmlpath;
            Properties.Settings.Default.Save();
            InitOsiContext();
        }
        private void InitOsiContext()
        {
            LoadContext();
            SetupFileWatcher();
        }
        internal void LoadContext()
        {
            XmlSerializer xser = new XmlSerializer();
            CurrentContext = xser.DeSerialize(_xmlpath, new XmlContext());
            _lastSyncTime = DateTime.Now;
        }
        private void SetupFileWatcher()
        {
            _xmlFileWatcher = new FileSystemWatcher(Path.GetDirectoryName(_xmlpath));
            _xmlFileWatcher.Changed += DataFileChanged;
            _xmlFileWatcher.Filter = Path.GetDirectoryName(_xmlpath);
            _xmlFileWatcher.EnableRaisingEvents = true;
        }
        private void DataFileChanged(object sender, FileSystemEventArgs e)
        {
            if (e.ChangeType == WatcherChangeTypes.Changed)
            {
                DataFileWasChanged?.Invoke();
            }
        }
        #region "Die Properties welche die Tabellen representieren"
        public ObservableCollection<ProductionLine> ProductionLines => CurrentContext.ProductionLines;
        public ObservableCollection<Instruction> Instructions => CurrentContext.Instructions;
        public ObservableCollection<MainStep> MainSteps => CurrentContext.MainSteps;
        public ObservableCollection<ChildStep> ChildSteps => CurrentContext.ChildSteps;
        #endregion
        public bool IsDataUpToDate
        {
            get { return new FileInfo(_xmlpath).LastWriteTime <= _lastSyncTime; }
        }
        public bool HasPessimisticLock
        {
            get { return File.Exists($"{Path.GetDirectoryName(_xmlpath)}\\{LockFileName}"); }
        }
        public bool SaveChanges()
        {
            if (HasPessimisticLock) ErrorMessageSent?.Invoke("Die pessimistische Sperre wurde verhängt, so ist ein speichern nicht möglich.");
            try
            {
                _xmlFileWatcher.EnableRaisingEvents = false;
                LockForChanges();
                XmlSerializer ser = new XmlSerializer();
                ser.Serialize<XmlContext>(_xmlpath, CurrentContext);
                _lastSyncTime = DateTime.Now;
                return true;
            }
            catch (Exception ex)
            {
                ErrorMessageSent?.Invoke(ex.Message);
                return false;
            }
            finally
            {
                UnlockForChanges();
                _xmlFileWatcher.EnableRaisingEvents = true;
            }
        }
        public void LockForChanges()
        {
            using (FileStream fs = new FileStream($@"{Path.GetDirectoryName(_xmlpath)}\{LockFileName}", FileMode.Create))
            {
                fs.Close();
            }
        }
        public void UnlockForChanges()
        {
            File.Delete($@"{Path.GetDirectoryName(_xmlpath)}\{LockFileName}");
        }
        internal void Seed()
        {
            LockForChanges();
            if(!ProductionLines.Any())
            {
                ProductionLines.Add(new ProductionLine() {Designation = "BioCel" });
                ProductionLines.Add(new ProductionLine() { Designation = "Flachmembran" });
                ProductionLines.Add(new ProductionLine() { Designation = "Kapillar" });
                ProductionLines.Add(new ProductionLine() { Designation = "Rohrmodule" });
            }
            if (!Instructions.Any())
            {
                Instructions.Add(new Instruction() {FormularNumber = 1000,FormularName = "TestFormular",RevisionNumber = 1,RevisionDate = DateTime.Now,LineID = ProductionLines.First().ID });
            }
            if(!MainSteps.Any())
            {
                MainSteps.Add(new MainStep() { StepNo = 1,StepText = "1. Main Step",StepImage = null,ImageText = "Image Text von Main Step 1",InstructionID = Instructions.First().ID });
            }
            if(!ChildSteps.Any())
            {
                ChildSteps.Add(new ChildStep() {StepNo = 1, StepText = "1. Child Step", StepImage = null, ImageText = "Image Text von Child Step 1", MainStepID = MainSteps.First().ID});
            }
            UnlockForChanges();
            SaveChanges();
        }

        private bool _disposedValue;
        // IDisposable
        protected virtual void Dispose(bool disposing)
        {
            if (!_disposedValue)
            {
                if (disposing)
                {
                    UnlockForChanges();
                    CurrentContext = null;
                    _xmlFileWatcher.Changed -= DataFileChanged;
                }
            }
            _disposedValue = true;
        }
        public void Dispose()
        {
            Dispose(true);
        }
    }
    [Serializable]
    public class XmlContext:NotifyBase
    {
        public XmlContext()
        {
            ProductionLines = new ObservableCollection<ProductionLine>();
            Instructions = new ObservableCollection<Instruction>();
            MainSteps = new ObservableCollection<MainStep>();
            ChildSteps = new ObservableCollection<ChildStep>();
        }
        private ObservableCollection<ProductionLine> _allProductionLines;
        public ObservableCollection<ProductionLine> ProductionLines
        {
            get { return _allProductionLines; }
            set { _allProductionLines = value;RaisePropertyChanged(); }
        }
        private ObservableCollection<Instruction> instructions;
        public ObservableCollection<Instruction> Instructions
        {
            get { return instructions; }
            set { instructions = value;RaisePropertyChanged(); }
        }
        private ObservableCollection<MainStep> mainSteps;
        public ObservableCollection<MainStep> MainSteps
        {
            get { return mainSteps; }
            set { mainSteps = value;RaisePropertyChanged(); }
        }
        private ObservableCollection<ChildStep> childSteps;
        public ObservableCollection<ChildStep> ChildSteps
        {
            get { return childSteps; }
            set { childSteps = value; RaisePropertyChanged(); }
        }
    }
}
